"""
Contains all teapot unpackers logic.
"""

from teapot.unpackers.unpacker import Unpacker
from teapot.unpackers.unpacker import register_unpacker
from teapot.unpackers.tarball_unpacker import TarballUnpacker
from teapot.unpackers.zipfile_unpacker import ZipFileUnpacker
